from wizzi_utils.socket import socket_tools as st
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.json import json_tools as jt
import socket
import os
import threading
from datetime import datetime

SERVER_ADDRESS = ('localhost', 10000)
BUF_LEN = 20
END_MSG = "$#$#"


def connect_to_server():
    connect_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_address = SERVER_ADDRESS
    print("\tClientOutput:Connecting to server {}".format(server_address))
    try:
        connect_socket.connect(server_address)
        print('\t\tClientOutput:Connected to {}'.format(server_address))
    except ConnectionRefusedError:
        assert False, '\t\tClientOutput:No server is found on {}'.format(server_address)

    j_out = {'name': 'client', 'msg': 'hello', 'time': 'msg 1'}
    j_out_str = jt.json_to_string(j_out)

    st.send_msg(
        connection=connect_socket,
        buflen=BUF_LEN,
        data=j_out_str,
        msg_end=END_MSG
    )

    print(
        st.buffer_to_str(
            data=j_out_str,
            prefix='ClientOutput:OUT',
            tabs=2
        )
    )

    j_in_str = st.receive_msg(connect_socket, buflen=BUF_LEN, msg_end=END_MSG)
    if j_in_str:
        print(
            st.buffer_to_str(
                data=j_in_str,
                prefix='ClientOutput:IN',
                tabs=2
            )
        )

        j_out = {'name': 'client', 'msg': 'hello', 'time': 'msg 3'}
        j_out_str = jt.json_to_string(j_out)

        st.send_msg(
            connection=connect_socket,
            buflen=BUF_LEN,
            data=j_out_str,
            msg_end=END_MSG
        )

        print(
            st.buffer_to_str(
                data=j_out_str,
                prefix='ClientOutput:OUT',
                tabs=2
            )
        )
    else:
        print('\t\tClientOutput:No Data from {}'.format(connect_socket))
    return


def open_server_test():
    mt.get_function_name(ack=True, tabs=0)
    sock = st.open_server(
        server_address=SERVER_ADDRESS,
        ack=True,
        tabs=1
    )

    # OPEN WITH A DIFFERENT THREAD THE CLIENT
    thread = threading.Thread(target=connect_to_server)
    thread.start()

    print('\t\tWaiting for connection {}/{}:'.format(1, 1))
    client_sock, client_address = sock.accept()
    j_in_str = st.receive_msg(client_sock, buflen=BUF_LEN, msg_end=END_MSG)
    if j_in_str:
        print(
            st.buffer_to_str(
                data=j_in_str,
                prefix='IN',
                tabs=2
            )
        )

        j_out = {'name': 'server', 'msg': 'wait', 't': 'msg 2'}
        j_out_str = jt.json_to_string(j_out)
        st.send_msg(
            connection=client_sock,
            buflen=BUF_LEN,
            data=j_out_str,
            msg_end="$#$#"
        )
        print(
            st.buffer_to_str(
                data=j_out_str,
                prefix='OUT',
                tabs=2
            )
        )

        j_in_str = st.receive_msg(client_sock, buflen=BUF_LEN, msg_end=END_MSG)
        if j_in_str:
            print(
                st.buffer_to_str(
                    data=j_in_str,
                    prefix='IN',
                    tabs=2
                )
            )
        else:
            print('\t\tNo Data from {}'.format(client_address))

        # CLIENT is wait for more messages
        # when finished - close client connection

        print('\t\tTerminating connection...')
        client_sock.close()
    else:
        print('\t\tNo Data from {}'.format(client_address))
    return


def get_host_name_test():
    mt.get_function_name(ack=True, tabs=0)
    print('\t{}'.format(st.get_host_name()))
    return


def get_active_ipv4_test():
    mt.get_function_name(ack=True, tabs=0)
    print('\tactive_ipv4: {}'.format(st.get_active_ipv4()))
    return


def buffer_to_str_test():
    mt.get_function_name(ack=True, tabs=0)
    data = 'hi server, how you doing???'  # len(data)==27
    print(st.buffer_to_str(data, prefix='client1', tabs=1, max_chars=27))
    print(st.buffer_to_str(data, prefix='client1', tabs=1, max_chars=26))
    print(st.buffer_to_str(data, prefix='client1', tabs=1, max_chars=15))
    return


def download_file_test():
    mt.get_function_name(ack=True, tabs=0)
    dir_path = mtt.TEMP_FOLDER1
    if not os.path.exists(dir_path):
        mt.create_dir(dir_path)
    dst = '{}/{}.jpg'.format(mtt.TEMP_FOLDER1, mtt.SO_LOGO)
    url = mtt.IMAGES_D[mtt.SO_LOGO]
    ret = st.download_file(url=url, dst_path=dst)
    if not ret:
        print('\tdownload of {} failed'.format(url))
    else:
        print('\t{} exists ? {}'.format(dst, os.path.exists(dst)))
        # st.download_file(url=mtt.IMAGES_D[mtt.SO_LOGO], dst_path=dst)  # check no overwrite - will fail
        mt.delete_dir_with_files(dir_path=mtt.TEMP_FOLDER1)
    return


def get_file_size_by_url_test():
    mt.get_function_name(ack=True, tabs=0)

    urls = [
        mtt.IMAGES_D[mtt.SO_LOGO],
        mtt.IMAGES_D[mtt.KITE],
        mtt.IMAGES_D[mtt.GIRAFFE],
        'https://cdn.sstatic.net/Sites/stackoverflow/img/NoSuchFile.png'
    ]
    for url in urls:
        url_file = url.split('/')[-1]
        s = st.get_file_size_by_url(url=url)
        if s == 'N/A':
            print('\tdownload of {} failed'.format(url_file))
        else:
            print('\tfile {} size is {}'.format(url_file, s))
    return


def get_wifi_pass_test():
    if mt.is_windows():
        mt.get_function_name(ack=True, tabs=0)
        d = st.get_wifi_passwords(ack=True)
        print('\tdict returned: {}'.format(d))
    return


def map_devices_test():
    mt.get_function_name(ack=True, tabs=0)
    # st.map_devices_arp(known_devices=None, ack=True)

    # if you know macs of devices
    known_devices = {
        'F8:1A:67:99:15:E0': 'my super cool gateway at home'
    }
    st.map_devices_arp(
        known_devices=known_devices, ack=True
    )
    return


def get_all_network_info_test():
    mt.get_function_name(ack=True, tabs=0)
    st.get_all_network_info(with_colons=True, ack=True)
    # st.get_all_network_info(with_colons=False, ack=True)
    return


def get_all_ipv4_test():
    mt.get_function_name(ack=True, tabs=0)
    st.get_wifi_ipv4(ack=True)
    st.get_ethernet_ipv4(ack=True)
    return


def get_all_macs_test():
    mt.get_function_name(ack=True, tabs=0)
    st.get_wifi_mac(ack=True)
    st.get_ethernet_mac(ack=True)
    return


def get_active_con_mac_test():
    mt.get_function_name(ack=True, tabs=0)
    st.get_active_con_mac(with_colons=True, ack=True)
    return


def get_mac_address_uuid_test():
    mt.get_function_name(ack=True, tabs=0)
    st.get_mac_address_uuid(ack=True, tabs=1)
    st.get_mac_address_uuid(with_colons=False, ack=True, tabs=1)
    print('\tMac address is {}'.format(st.get_mac_address_uuid()))
    return


# noinspection PyPackageRequirements
def big_query_package_by_month_test():
    """
    # package first release Feb 22, 2021 so not expecting to have 2021-01-01
    # Run started at 27-10-2022 09:36:49
    results for last 21 months:
        pip_reqs       month
    0        149  2022-10-01
    1         22  2022-09-01
    2         23  2022-08-01
    3         56  2022-07-01
    4         49  2022-06-01
    5         42  2022-05-01
    6         21  2022-04-01
    7         64  2022-03-01
    8         16  2022-02-01
    9          1  2022-01-01
    10         1  2021-12-01
    11        24  2021-11-01
    12        27  2021-10-01
    13         2  2021-09-01
    14        53  2021-08-01
    15        83  2021-07-01
    16        78  2021-06-01
    17        91  2021-05-01
    18        78  2021-04-01
    19        38  2021-03-01
    20         6  2021-02-01
    time for all: 0:03:30


    results for last 21 months:
         hits       month
    0    3214  2022-10-01
    1    2270  2022-09-01
    2    2916  2022-08-01
    3    3297  2022-07-01
    4    3679  2022-06-01
    5    2651  2022-05-01
    6    2647  2022-04-01
    7    5968  2022-03-01
    8    3846  2022-02-01
    9    3439  2022-01-01
    10   2484  2021-12-01
    11   3065  2021-11-01
    12   4181  2021-10-01
    13   3443  2021-09-01
    14   5326  2021-08-01
    15  12155  2021-07-01
    16   7951  2021-06-01
    17  13929  2021-05-01
    18   6436  2021-04-01
    19   1748  2021-03-01
    20    231  2021-02-01
    time for all: 0:03:32
    """

    mt.get_function_name(ack=True, tabs=0)

    cred_file = '{}/secret_keys/credentialsBigQ.json'.format(mt.get_repo_root(repo_name='wizzi_utils'))
    if not os.path.exists(cred_file):
        mt.exception_error(e='credentials file not found {}'.format(os.path.abspath(cred_file)))
        return

    mt.set_env_variable(
        key='GOOGLE_APPLICATION_CREDENTIALS',
        val=cred_file,
        ack=True
    )
    months_from_wu_release = 3
    # def diff_month(d1: datetime, d2: datetime) -> int:
    #     return (d1.year - d2.year) * 12 + d1.month - d2.month
    # WU_PUBLISH_DATE = datetime(year=2021, month=2, day=22)
    # months_from_wu_release = diff_month(datetime.today(), WU_PUBLISH_DATE) + 1  # +1: include this month

    st.big_query_package_by_month(
        package_name='wizzi-utils',  # package first release Feb 22, 2021
        months_back=months_from_wu_release,
        only_pip=True
    )

    st.big_query_package_by_month(
        package_name='wizzi-utils',  # package first release Feb 22, 2021
        months_back=months_from_wu_release,
        only_pip=False
    )
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    open_server_test()
    get_host_name_test()
    get_active_ipv4_test()
    buffer_to_str_test()
    download_file_test()
    get_file_size_by_url_test()
    get_wifi_pass_test()
    map_devices_test()
    get_all_network_info_test()
    get_all_ipv4_test()
    get_all_macs_test()
    get_active_con_mac_test()
    get_mac_address_uuid_test()
    big_query_package_by_month_test()
    print('{}'.format('-' * 20))
    return



"""
for tflite_runtime:
pip install --index-url https://google-coral.github.io/py-repo/ tflite_runtime # windows 10
"""
try:
    from wizzi_utils.tflite.tflite_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.tflite import test



"""
for torch 181 gpu cuda 101:
pip install torch==1.8.1+cu101 torchvision==0.9.1+cu101 torchaudio==0.8.1 -f https://download.pytorch.org/whl/torch_stable.html
for torch 182 gpu cuda 102:
pip install torch==1.8.2 torchvision==0.9.2 torchaudio===0.8.2 --extra-index-url https://download.pytorch.org/whl/lts/1.8/cu102
"""
try:
    from wizzi_utils.torch.torch_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.torch import test



from wizzi_utils import misc_tools as mt  # misc tools
# noinspection PyPackageRequirements
import tflite_runtime


# from tflite_runtime.interpreter import Interpreter

# try:
#     # noinspection PyUnresolvedReferences
#     interpreter_found = tflite_runtime.interpreter.Interpreter
#     print('success using tflite_runtime.interpreter.Interpreter')
# except AttributeError:
#     # print('failure using tflite_runtime.interpreter.Interpreter')
#     try:
#         # noinspection PyPackageRequirements,PyUnresolvedReferences
#         import tensorflow as tf
#
#         interpreter_found = tf.lite.Interpreter
#         print('success using tf.lite.Interpreter')
#     except (ImportError, AttributeError):
#         pass
#         print('failure using tf.lite.Interpreter')


def get_tflite_version(ack: bool = False, tabs: int = 1) -> str:
    """
    :param ack:
    :param tabs:
    :return:
    see get_tflite_version_test()
    """
    string = mt.add_color('{}* TFLite Version {}'.format(tabs * '\t', tflite_runtime.__version__), ops=mt.SUCCESS_C)
    # string += mt.add_color(' - GPU detected ? ', op1=mt.SUCCESS_C)
    # if gpu_detected():
    #     string += mt.add_color('True', op1=mt.SUCCESS_C2[0], extra_ops=mt.SUCCESS_C2[1])
    # else:
    #     string += mt.add_color('False', op1=mt.FAIL_C2[0], extra_ops=mt.FAIL_C2[1])
    if ack:
        print(string)
    return string


# def get_tensorflow_version(ack: bool = False, tabs: int = 1) -> str:
#     """
#     :param ack:
#     :param tabs:
#     :return:
#     see get_tensorflow_version_test()
#     """
#     import os
#     import warnings
#
#     os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # removes tf informative messages
#     warnings.simplefilter(action='ignore', category=FutureWarning)  # remove tf future warnings
#     # noinspection PyPackageRequirements,PyUnresolvedReferences
#     import tensorflow as tf  # noqa E402
#     # print(tf.__version__)
#     # print(tf.VERSION)
#     # print(tf.version.VERSION)
#     # noinspection PyUnresolvedReferences
#     string = mt.add_color('{}* TensorFlow Version {}'.format(tabs * '\t', tf.__version__), ops=mt.SUCCESS_C)
#     string += mt.add_color(' - GPU detected ? ', ops=mt.SUCCESS_C)
#     if gpu_detected():
#         string += mt.add_color('True', ops=mt.SUCCESS_C2)
#     else:
#         string += mt.add_color('False', ops=mt.FAIL_C2)
#     if ack:
#         print(string)
#     return string
#
#
# def gpu_detected() -> bool:
#     """
#     :return:
#     """
#     gpu_on = False
#     try:  # version >= 2
#         # noinspection PyUnresolvedReferences
#         if len(tf.config.list_physical_devices('GPU')) >= 1:
#             gpu_on = True
#     except AttributeError:
#         # noinspection PyUnresolvedReferences
#         if 'GPU' in tf.test.gpu_device_name():
#             gpu_on = True
#     return gpu_on



import sys
import os
import subprocess
import inspect
# noinspection PyPackageRequirements
import pyttsx3  # pip install pyttsx3


# @staticmethod
# def onStart(name):
#     print('starting', name)
#
# @staticmethod
# def onWord(name, location, length):
#     print('word', name, location, length)
#
# @staticmethod
# def onEnd(name, completed):
#     print('finishing', name, completed)

# self.buddy = pyttsx3.init()
# self.buddy.connect('started-utterance', self.onStart)
# self.buddy.connect('started-word', self.onWord)
# self.buddy.connect('finished-utterance', self.onEnd)

def get_file_name() -> str:
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[1]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.filename)
    except IndexError as e:
        print(e)
    return ret_val


class MachineBuddy:
    """
    text to speech
    https://pypi.org/project/pyttsx3/
    """

    def __init__(self, voice_ind: int = 0, rate: int = 200, vol: float = 1.0) -> None:
        """
        :param voice_ind: default voice - num 0 in self.buddy.getProperty('voices')
        :param rate: default rate 200
        :param vol: default vol 1.0 (from 0.0 to 1.0)
        """
        self.buddy = pyttsx3.init()
        # noinspection PyTypeChecker
        self.voices_num = len(self.buddy.getProperty('voices'))
        self.current_voice_ind = 0
        if 0 < voice_ind <= self.voices_num - 1:  # 0 by default
            self.change_voice(voice_ind)
        if 0 < rate != 200:
            self.change_rate(rate)
        if 0.0 <= vol < 1.0:
            self.change_volume(vol)
        self.buddy.runAndWait()
        return

    def __del__(self) -> None:
        # if self.buddy is not None:
        #     self.buddy.stop()
        return

    def __str__(self) -> str:
        string = 'MachineBuddy:\n'
        string += '\tvoice({}):{}\n'.format(self.get_current_property(key='voice_idx'),
                                            self.get_current_property(key='voice_name'))
        string += '\tvol:{}\n'.format(self.get_current_property(key='volume'))
        string += '\trate:{}'.format(self.get_current_property(key='rate'))
        return string

    def get_all_voices(self, ack: bool = True, tabs: int = 1):
        """
        :param ack:
        :param tabs:
        :return:
        TO add more speakers on Windows10:
        based on https://www.youtube.com/watch?v=KMtLqPi2wiU&ab_channel=MuruganS
        step 1(download):
            ctrl+winKey+n - open narrator
            add more voices
            add voice
            select, install and wait for it to finish
            # the new voice available only to Microsoft
            i will refer to this language as _NEW_LANGUAGE
        step 2(make the installed voice available to all tts programs):
            start, regedit, open as administrator:
                goto: Computer/HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens
                click export on any language and save to desktop(we just need a path) -> i will call it f1
                goto: Computer/HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE
                select the _NEW_LANGUAGE installed on step 1. export to desktop -> i will call it f2
                open both saved files on notepad++
                from f1 - copy the paths
                e.g.
                path1: [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/TTS_MS_EN-GB_HAZEL_11.0]
                path2: [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/TTS_MS_EN-GB_HAZEL_11.0/Attributes]
                we care only for the dir name:
                so path1: [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/SOME_LANGUAGE]
                so path2: [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/SOME_LANGUAGE/Attributes]

                open f2
                duplicate the first 2 paragraphs (first is the actual language and second is the attributes)
                paragraph 1:
                    [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE]
                paragraph 2:
                    [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE/Attributes]
                now you have 4 paragraphs instead of 2:
                p1 [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE]
                p2 [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE/Attributes]
                p3 [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE]
                p4 [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/_NEW_LANGUAGE/Attributes]
                change p3 to path1 and change SOME_LANGUAGE to _NEW_LANGUAGE
                change p4 to path2 and change SOME_LANGUAGE to _NEW_LANGUAGE
                e.g.
                from
                [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/MSTTS_V110_heIL_Asaf]
                to
                [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/MSTTS_V110_heIL_Asaf]
                and from
                [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech_OneCore/Voices/Tokens/MSTTS_V110_heIL_Asaf/Attributes]
                to
                [HKEY_LOCAL_MACHINE/SOFTWARE/Microsoft/Speech/Voices/Tokens/MSTTS_V110_heIL_Asaf/Attributes]
                save and exit notepad++
                run the saved file and _NEW_LANGUAGE should appear in any tts program
        """

        # all_voice = self.buddy.getProperty('voices')
        all_voices = self.get_current_property('all_voices')
        if ack:
            # noinspection PyTypeChecker
            for v in all_voices:
                print('{}{}'.format(tabs * '\t', v))
        return all_voices

    def change_voice(self, new_voice_ind: int) -> None:
        """
        :param new_voice_ind:
        voices differ between computers
        :return:
        """
        if 0 <= new_voice_ind <= self.voices_num - 1:
            all_voices = self.get_current_property('all_voices')
            # noinspection PyUnresolvedReferences
            self.buddy.setProperty(name='voice', value=all_voices[new_voice_ind].id)
            self.current_voice_ind = new_voice_ind
            self.buddy.runAndWait()
        return

    def change_rate(self, new_rate: int) -> None:
        """
        default rate 200
        :param new_rate:
        :return:
        """
        if new_rate > 0:
            self.buddy.setProperty('rate', new_rate)  # setting up new voice rate
            self.buddy.runAndWait()
        return

    def change_volume(self, new_vol: float) -> None:
        """
        default 1.0
        :param new_vol:
        :return:
        setting up volume level between 0 and 1
        """
        if 0.0 <= new_vol <= 1.0:  # 1 by default
            self.buddy.setProperty('volume', new_vol)
            self.buddy.runAndWait()
        return

    def say(self, text: str) -> None:
        """
        :param text:
        :return:
        """
        self.buddy.say(text)
        self.buddy.runAndWait()
        return

    @staticmethod
    def speak(text: str, block: bool = True) -> None:
        """
        :param text:
        :param block: if True - program waits till speaking is over
            if False - run the speak function with another process
        :return:
        static use with default params
        """
        if block:
            pyttsx3.speak(text)
        else:
            tts_path = get_file_name()
            # print(tts_path)
            # print(os.path.exists(tts_path))
            # subprocess.call([r"C:\Users\GiladEiniKbyLake\.conda\envs\wu\python.exe", "tts.py", phrase])
            # subprocess.run([r"C:\Users\GiladEiniKbyLake\.conda\envs\wu\python.exe", "tts.py", phrase])
            subprocess.Popen([sys.executable, tts_path, text])  # Call subprocess
            # args = [tts_path, '--text', text]
            # p = subprocess.Popen([sys.executable or 'python'] + args)
            # p.wait()
        return

    def get_current_property(self, key):
        val = None
        if key == 'voice_idx':
            val = self.current_voice_ind
        elif key == 'voice_name':
            voice = self.buddy.getProperty('voice')
            # noinspection PyTypeChecker
            val = os.path.basename(voice)
        elif key == 'rate':
            val = self.buddy.getProperty('rate')
        elif key == 'volume':
            val = self.buddy.getProperty('volume')
        elif key == 'all_voices':
            val = self.buddy.getProperty('voices')
        return val


try:
    # noinspection PyPackageRequirements
    from PyQt5 import QtCore, QtGui, QtWidgets  # pip install PyQt5
    from wizzi_utils.pyplot import pyplot_tools as pyplt

    # noinspection PyTypeChecker,PyUnresolvedReferences
    class MachineBuddyGui(QtWidgets.QMainWindow):
        def __init__(self, voice_idx: int = 0, rate: int = 200, volume: int = 100, loc: str = 'bl',
                     common_sentences: list = None):
            """
            :param voice_idx: index of voice from the voices installed on your device
                see machine_buddy.get_all_voices(ack=True)
            :param rate: voice rate
            :param volume: voice volume from 0 to 100
            :param loc: bl -> bottom left
            :param common_sentences: list of sentences. will be added in a combo box for easy access
            """
            super().__init__()

            self.machine_buddy = MachineBuddy(
                voice_ind=voice_idx,
                rate=rate,
                vol=volume / 100
            )

            # get the values from machine_buddy incase an illegal value was inserted
            voice_idx = self.machine_buddy.get_current_property(key='voice_idx')
            volume = int(self.machine_buddy.get_current_property(key='volume') * 100)
            rate = self.machine_buddy.get_current_property(key='rate')

            # self.app = QtWidgets.QApplication(sys.argv)
            self.setWindowTitle('MachineBuddyGui')
            w, h = 640, 215
            self.resize(w, h)
            self.move_window(self, win_loc=loc, fig_w=w, fig_h=h)
            grid = QtWidgets.QGridLayout()
            # grid.setHorizontalSpacing(0)
            # grid.setVerticalSpacing(0)
            # grid.setSpacing(0)

            # widget.setFocusPolicy(Qt.NoFocus)

            row = 0
            self.cur_settings_title_label = QtWidgets.QLabel('Current settings:')
            self.cur_settings_voice_fmt = 'voice({})={}'
            self.cur_settings_voice = QtWidgets.QLabel('')  # will be set after voice str calculated
            self.cur_settings_voice.setStyleSheet('border: 1px solid red')
            self.cur_settings_rate_vol_fmt = 'rate={}, vol={}%'
            self.cur_settings_rate_vol = QtWidgets.QLabel(self.cur_settings_rate_vol_fmt.format(rate, volume))
            self.cur_settings_rate_vol.setStyleSheet('border: 1px solid red')
            grid.addWidget(self.cur_settings_title_label, row, 0)
            grid.addWidget(self.cur_settings_voice, row, 1)
            grid.addWidget(self.cur_settings_rate_vol, row, 2)

            row = 1
            self.voices_label = QtWidgets.QLabel('Voices:')
            # self.voices_label.setStyleSheet('border: 1px solid red')
            self.voices_combo = QtWidgets.QComboBox()
            all_voices = self.machine_buddy.get_all_voices(ack=False)
            self.all_voices_name_to_idx = {}
            for i, v in enumerate(all_voices):
                self.all_voices_name_to_idx[v.name] = i
            self.voices_combo.addItems(list(self.all_voices_name_to_idx.keys()))
            self.voices_combo.setCurrentIndex(voice_idx)
            self.voices_update = QtWidgets.QPushButton('Update')
            self.voices_update.clicked.connect(self.update_voice)
            grid.addWidget(self.voices_label, row, 0)
            grid.addWidget(self.voices_combo, row, 1)
            grid.addWidget(self.voices_update, row, 2)

            # set current setting active voice
            self.cur_settings_voice.setText(
                self.cur_settings_voice_fmt.format(voice_idx, self.voices_combo.currentText()))

            row = 2
            self.volume_label = QtWidgets.QLabel('Volume {:3}%'.format(volume))
            self.volume_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
            # self.volume_slider.setMaximumWidth(1000)
            # self.volume_slider.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)
            self.volume_slider.setRange(0, 100)
            self.volume_slider.setValue(volume)
            self.volume_slider.valueChanged.connect(self.update_volume_label)
            self.volume_update = QtWidgets.QPushButton('Update')
            self.volume_update.clicked.connect(self.update_volume)
            grid.addWidget(self.volume_label, row, 0)
            grid.addWidget(self.volume_slider, row, 1)
            grid.addWidget(self.volume_update, row, 2)

            row = 3
            self.rate_label = QtWidgets.QLabel('rate:')
            self.rate_line_edit = QtWidgets.QLineEdit()
            self.rate_line_edit.setText('{}'.format(rate))
            self.rate_update = QtWidgets.QPushButton('Update')
            self.rate_update.clicked.connect(self.update_rate)
            grid.addWidget(self.rate_label, row, 0)
            grid.addWidget(self.rate_line_edit, row, 1)
            grid.addWidget(self.rate_update, row, 2)

            row = 4
            self.text_to_say_label = QtWidgets.QLabel('Text to say:')
            self.text_to_say_line_edit = QtWidgets.QLineEdit()
            self.text_to_say_line_edit.setText('Shalom')
            self.text_to_say_line_edit.returnPressed.connect(self.text_to_say_button_clicked)
            self.text_to_say_button = QtWidgets.QPushButton('Say')
            # self.text_to_say_speak.setFixedWidth(40)
            self.text_to_say_button.clicked.connect(self.text_to_say_button_clicked)
            grid.addWidget(self.text_to_say_label, row, 0)
            grid.addWidget(self.text_to_say_line_edit, row, 1)
            grid.addWidget(self.text_to_say_button, row, 2)

            row = 5
            self.say_again_label = QtWidgets.QLabel('Last sentence:')
            self.say_again_line_edit = QtWidgets.QLineEdit()
            self.say_again_line_edit.setText('')
            # self.say_again_line_edit.setDisabled(True)
            # self.say_again_line_edit.returnPressed.connect(self.say_again_clicked)
            self.say_again_button = QtWidgets.QPushButton('Say again')
            # self.text_to_say_speak.setFixedWidth(40)
            self.say_again_button.clicked.connect(self.say_again_button_clicked)
            grid.addWidget(self.say_again_label, row, 0)
            grid.addWidget(self.say_again_line_edit, row, 1)
            grid.addWidget(self.say_again_button, row, 2)

            row = 6
            self.common_label = QtWidgets.QLabel('Common:')
            self.common_combo = QtWidgets.QComboBox()
            common_base = [
                'Hello',
            ]
            self.common_combo.addItems(common_base)
            if common_sentences is not None:
                self.common_combo.addItems(common_sentences)
            self.common_say_button = QtWidgets.QPushButton('Say')
            self.common_say_button.clicked.connect(self.common_say_button_clicked)

            grid.addWidget(self.common_label, row, 0)
            grid.addWidget(self.common_combo, row, 1)
            # grid.addWidget(self.common_combo, row, 1, 1, 2)  # rowspan 1 and colspan 2
            grid.addWidget(self.common_say_button, row, 2)

            main_widget = QtWidgets.QWidget()
            main_widget.setLayout(grid)
            self.setCentralWidget(main_widget)
            self.show()

            return

        def update_voice(self):
            new_voice = self.voices_combo.currentText()
            new_voice_idx = self.all_voices_name_to_idx[new_voice]
            self.machine_buddy.change_voice(new_voice_ind=new_voice_idx)
            print('voice changed to {} idx {}'.format(new_voice, new_voice_idx))
            self.cur_settings_voice.setText(  # current voice
                self.cur_settings_voice_fmt.format(
                    new_voice_idx, new_voice
                )
            )
            return

        def update_volume_label(self):
            self.volume_label.setText('Volume {:03}%'.format(self.volume_slider.value()))
            return

        def update_volume(self):
            new_vol = self.volume_slider.value() / 100
            self.machine_buddy.change_volume(new_vol=new_vol)
            print('volume changed to {}'.format(new_vol))
            self.cur_settings_rate_vol.setText(
                self.cur_settings_rate_vol_fmt.format(
                    self.machine_buddy.get_current_property(key='rate'), self.volume_slider.value())
            )
            return

        def update_rate(self):
            new_rate = self.rate_line_edit.text()
            if new_rate.isnumeric():
                new_rate = int(new_rate)
                self.machine_buddy.change_rate(new_rate=new_rate)
                print('rate changed to {}'.format(new_rate))
                self.cur_settings_rate_vol.setText(
                    self.cur_settings_rate_vol_fmt.format(
                        new_rate, int(self.machine_buddy.get_current_property(key='volume') * 100))
                )
            else:
                print('value {} must be int'.format(new_rate))
            return

        def text_to_say_button_clicked(self):
            text_to_say = self.text_to_say_line_edit.text()
            if text_to_say:
                self.text_to_say_line_edit.clear()
                print('saying {}'.format(text_to_say))
                self.machine_buddy.say(text=text_to_say)
                self.say_again_line_edit.setText(text_to_say)
            return

        def say_again_button_clicked(self):
            text_to_say_again = self.say_again_line_edit.text()
            if text_to_say_again:
                print('saying again {}'.format(text_to_say_again))
                self.machine_buddy.say(text=text_to_say_again)
            return

        def common_say_button_clicked(self):
            common_say = self.common_combo.currentText()
            if common_say:
                print('common say {}'.format(common_say))
                self.machine_buddy.say(text=common_say)
            return

        def keyPressEvent(self, event: QtGui.QKeyEvent):
            if event.key() in [QtCore.Qt.Key_Escape]:
                print('keyPressEvent(): Qt plot: Esc was clicked. Terminating...')
                MachineBuddyGui.get_qt_app().quit()

        @staticmethod
        def get_qt_app() -> QtWidgets.QApplication:
            app = QtWidgets.QApplication.instance()  # main app must init before creating GLViewWidget
            if QtWidgets.QApplication.instance() is None:
                app = QtWidgets.QApplication(sys.argv)
            return app

        def run(self):
            return sys.exit(self.app.exec_())

        @staticmethod
        def move_window(widget, win_loc, fig_w, fig_h):
            # move window
            window_w, window_h = pyplt.screen_dims()  # screen dims in pixels
            taskbar_size = 70
            x, y = None, None
            if win_loc == 'tl':  # top left
                x, y = 0, 0
            elif win_loc == 'tc':  # top center
                x, y = (window_w - fig_w) / 2, 0
            elif win_loc == 'tr':  # top right
                x, y = window_w - fig_w, 0
            elif win_loc == 'cl':  # center left
                x, y = 0, (window_h - fig_h) / 2
            elif win_loc == 'cc':  # center center
                x, y = (window_w - fig_w) / 2, (window_h - fig_h) / 2
            elif win_loc == 'cr':  # center right
                x, y = window_w - fig_w, (window_h - fig_h) / 2 - taskbar_size
            elif win_loc == 'bl':  # bottom left
                x, y = 0, window_h - fig_h - taskbar_size
            elif win_loc == 'bc':  # bottom center
                x, y = (window_w - fig_w) / 2, window_h - fig_h - taskbar_size
            elif win_loc == 'br':  # bottom right
                x, y = window_w - fig_w, window_h - fig_h - taskbar_size
            widget.move(x, y)
            return


    def run_machine_buddy_gui(
            voice_idx: int = 0,
            rate: int = 200,
            volume: int = 100,
            loc: str = 'bl',
            common_sentences: list = None
    ) -> None:
        """see class MachineBuddyGui documentation"""
        app = MachineBuddyGui.get_qt_app()
        # noinspection PyUnusedLocal
        mbg = MachineBuddyGui(voice_idx=voice_idx, rate=rate, volume=volume, loc=loc, common_sentences=common_sentences)
        return sys.exit(app.exec_())

except (ModuleNotFoundError, ImportError):
    pass



import socket
import os
import threading
import psutil
from uuid import getnode as _get_mac_uuid
from urllib.request import urlretrieve, urlopen
from urllib.error import URLError
from datetime import datetime
from wizzi_utils.misc import misc_tools as mt


def open_server(server_address: tuple = ('localhost', 10000), ack: bool = True, tabs: int = 1) -> socket:
    """
    :param server_address:
    :param ack:
    :param tabs:
    see open_server_test()
    """
    if ack:
        print('{}Opening server on IP,PORT {}'.format(tabs * '\t', server_address))
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(server_address)
    sock.listen(1)
    return sock


def get_host_name() -> str:
    """
    :return: hostname
    try using misc_tools.get_pc_name() instead
    see get_host_name_test()
    """
    hostname = socket.gethostname()
    return hostname


def get_active_ipv4() -> str:
    """
    :return ipv4 address of this computer
    see get_ipv4_test()
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ipv4 = s.getsockname()[0]
    except OSError as e:
        mt.exception_error('No connection: {}'.format(e), real_exception=True)
        ipv4 = 'N/A'
    return ipv4


def send_msg(connection: socket, buflen: int, data: str, msg_end: str) -> None:
    """
    :param connection: the connection of this client/server to the server/client
    :param buflen: needed to split the message if it is longer than 'buflen'
    :param data: string to send
    :param msg_end: special string that notifies when the msg is over(faster than try catch)
        MUST BE A STRING THAT CANT APPEAR ON MESSAGES - e.g. "$#$#"
    see open_server_test()
    """
    data_e = str.encode(data + msg_end)
    data_e_len = len(data_e)
    for i in range(0, data_e_len, buflen):
        chunk_i = data_e[i:i + buflen]
        connection.send(chunk_i)
    return


def receive_msg(connection: socket, buflen: int, msg_end: str) -> str:
    """
    :param connection: the connection of this client/server to the server/client
    :param buflen: needed to receive the message in chunks
    :param msg_end: special string that notifies when the msg is over(faster than try catch)
        MUST BE A STRING THAT CANT APPEAR ON MESSAGES - e.g. "$#$#"
    :return: string of the received data
    see open_server_test()
    """
    data_in = ''
    saw_end_delimiter = False
    while not saw_end_delimiter:
        data_in += connection.recv(buflen).decode('utf-8')
        if not data_in:
            break  # empty transmission
        if data_in.endswith(msg_end):
            data_in = data_in.replace('$#$#', '')
            saw_end_delimiter = True

    return data_in


def buffer_to_str(data: str, prefix: str, tabs: int = 1, max_chars: int = 100) -> str:
    """
    :param data: data as string
    :param prefix: string prefix e.g. 'in', 'out', 'server', 'client'
    :param tabs:
    :param max_chars:
    :return: pretty print of the buffer
    see buffer_to_str_test()
    """
    data_len = len(data)
    data_chars = data_len + 1 if data_len <= max_chars else max_chars

    msg = '{}{}: {} (bytes sent={})'.format(tabs * '\t', prefix, data[:data_chars], data_len)
    if data_len > max_chars:
        msg += ' ... message is too long'
    return msg


def download_file(url: str, dst_path: str = './file', tabs: int = 1) -> bool:
    """
    :param url:
    :param dst_path: where to save the file.
        if dst_path contains non existing folders - it will fail.
            use mt.create_dir(dir) for new dirs
    :param tabs:
    :return: bool 1 for success
    see download_file_test()
    see load_img_from_web() in test_open_cv_tools.py
    """
    filename = url.split('/')[-1]

    def download_progress_hook(count, blockSize, totalSize):
        if totalSize < 0:  # saw this once
            percent = 'N/A'
            size_s = 'size N/A'
        else:
            percent = '{:.1f}%'.format(min(float(count * blockSize) / float(totalSize) * 100.0, 100))
            size_s = mt.convert_size(totalSize)
        print("\r{}Completed: {} from {} - {}".format((tabs + 1) * '\t', percent, size_s, filename), end="")
        return

    ret = False
    if not os.path.exists(dst_path):
        try:
            if mt.is_windows():
                # in windows - add prefix which allows the file name to be longer than the default len (70 chars)
                dst_path = mt.full_path_no_limit(dst_path)
                if len(os.path.basename(dst_path)) > 255:
                    err = 'len(os.path.basename(dst_path))={} exceeds the maximal length which is 255 chars. '
                    err += 'dst_path = {}'
                    mt.exception_error(err.format(len(os.path.basename(dst_path)), os.path.basename(dst_path)),
                                       real_exception=False, tabs=tabs)
                    return False
            msg = 'Downloading from {} to {}'.format(url, dst_path)
            print('{}{}'.format(tabs * '\t', mt.add_color(msg, ops='light_magenta')))
            urlretrieve(url, dst_path, download_progress_hook)
            print()  # release cartridge
            ret = True
        except URLError as e:
            mt.exception_error('{} {}'.format(e, url), real_exception=True, tabs=tabs + 1)
        except FileNotFoundError as e:
            mt.exception_error('{}'.format(e), real_exception=True, tabs=tabs + 1)

    else:
        mt.exception_error(mt.EXISTS.format(dst_path), real_exception=False, tabs=tabs)
    return ret


def get_file_size_by_url(url: str) -> str:
    try:
        obj_info = urlopen(url)
        size_int = int(obj_info.getheader('Content-Length'))
        size_pretty = mt.convert_size(size_bytes=size_int)
    except URLError as e:
        mt.exception_error('with file {} error: {}'.format(url, e), real_exception=True, tabs=1)
        size_pretty = 'N/A'
    return size_pretty


def get_wifi_password(profile_name: str, ack: bool = True) -> str:
    """
    get a wifi saved password on this computer of a specific profile
    :return:
    """
    cmd = 'netsh wlan show profile \"{}\" key=clear'  # wrap the name in-case it has spaces
    netsh_profile_out = mt.run_shell_command_and_get_out(cmd.format(profile_name), ack_cmd=False)
    password = 'Not Found'
    for i, line in enumerate(netsh_profile_out):
        # print('{}){}'.format(i, line))
        if 'Key Content' in line:
            password = line.split(':')[1].strip()
            # print(repr(password))
            break
    if ack:
        print('\t{:<20}| {}'.format(profile_name, password))
    return password


def get_wifi_passwords(ack: bool = True) -> dict:
    """
    get all wifi saved passwords on this computer
    :return:
    """
    profile_to_password_d = {}
    if mt.is_windows():
        # fetch all profiles
        netsh_out = mt.run_shell_command_and_get_out('netsh wlan show profiles', ack_cmd=False)
        profile_to_password_d = {}
        for i, line in enumerate(netsh_out):
            # print('}){}'.format(i, line))
            if 'All User Profile' in line:
                p_name = line.split(':')[1].strip()
                # print(repr(p_name))
                profile_to_password_d[p_name] = None

        # for each profile, fetch password
        if ack:
            print('Profiles and Passwords:')

        for p_name in profile_to_password_d:
            password = get_wifi_password(profile_name=p_name, ack=False)
            profile_to_password_d[p_name] = password
            if ack:
                print('\t{:<20}| {}'.format(p_name, password))
    return profile_to_password_d


def check_if_received(out_lines: list) -> bool:
    found = False
    for i, line in enumerate(out_lines):
        if 'Reply from' in line:
            # print('{}){}'.format(i, line))
            found = False if 'Destination host unreachable' in line else True
            break
    return found


def check_ip(ip: str) -> bool:
    # print('checking ip {}...'.format(ip))
    out_lines = mt.run_shell_command_and_get_out(cmd='ping -n 1 {}'.format(ip), ack_cmd=False, ack_out=False)
    found = check_if_received(out_lines)
    # if found:
    #     print('\tip {} is logged in'.format(ip))
    return found


def check_ip_thread(ip: str, ips_found: list) -> None:
    # print('checking ip {}...'.format(ip))
    out_lines = mt.run_shell_command_and_get_out(cmd='ping -n 1 {}'.format(ip), ack_cmd=False, ack_out=False)
    found = check_if_received(out_lines)
    if found:
        ips_found.append(ip)
        # print('\tip {} is logged in'.format(ip))
    return


def map_brute_force(start_idx: int = 0, end_idx: int = 254, ack: bool = True) -> list:
    """
    :param start_idx:
    :param end_idx:
    :param ack:
    :return: list of ips detected (replied to ping)
    e.g. output
    --------------------------------------------------------------------------------
    local network prefix 192.168.0
    checking from 192.168.0.X where X goes from 0 to 255...
        ip 192.168.0.2 is logged in
        ip 192.168.0.6 is logged in
        ip 192.168.0.8 is logged in
    --------------------------------------------------------------------------------
    Total run time 0:00:04
    """
    ips_found = []
    my_ip = get_active_ipv4()
    if my_ip == 'N/A':
        return ips_found
    pref_ip = my_ip.split('.')[:3]  # take first 3 chars e.g. 192.168.0.8 -> 192.168.0
    pref_ip = '.'.join(pref_ip)
    pref_ip += '.{}'  # add variable format
    if ack:
        print('local network prefix {}'.format(pref_ip))
        print('checking from {} where X goes from {} to {}...'.format(pref_ip.format('X'), start_idx, end_idx))
    pool = list()

    for i in range(start_idx, end_idx):
        address = pref_ip.format(i)
        # check_ip(ip=address)
        x = threading.Thread(target=check_ip_thread, args=(address, ips_found))
        pool.append(x)
        x.start()
    for index, thread in enumerate(pool):
        thread.join()
    ips_found.sort()
    if ack:
        for ip in ips_found:
            print('\tip {} is logged in'.format(ip))
    return ips_found


def map_devices_arp(known_devices: dict = None, ack: bool = True):
    """
    :param known_devices: if you know the mac of a device, give it a nickname and it will be displayed
        e.g.
        known_devices = {
            '70:4D:7B:8A:65:EE': 'Wizzi-Dorms (Ethernet adapter Ethernet)',
            'D4-6E-0E-16-51-41': 'Wizzi-Dorms (Wireless LAN adapter Wi-Fi)',
            'A2:8C:1E:42:57:64': 'Wizzi-Pixel3a',
            '5C:80:B6:30:1A:7E': 'Wizzi-Dell (Wireless LAN adapter Wi-Fi)',
        }
        notice Wizzi-Dorms is my home pc and have 2 mac
        cmd > ipconfig /all
        under Ethernet adapter Ethernet one mac and under Wireless LAN adapter Wi-Fi another
        one for lan and one for wifi
        on linux:
        shell > ifconfig
        wlan0 and eth0 are the ones we need
    :param ack:
    :return:
    TODO check on linux
    Notice that if you get something like:
    File "C:/Users/GiladEiniKbyLake/.conda/envs/bin_env/lib/subprocess.py", line 516,
        in run: Command 'ping -n 1 192.168.0.7' returned non-zero exit status 1.
    it's probably a firewall issue on that device.
    if you wish that device will be discovered and it uses windows firewall,
    on the device, you can add inbound rule to win defender:
        https://activedirectorypro.com/allow-ping-windows-firewall/
    control panel > firewall > advanced > inbound rules
    sort by group and go to group file and printer sharing
    select "File and Printer Sharing (Echo Request - ICMPv4-In)" (should be and entry to private and to any)
    change to enabled and done

    e.g.
    # if you know macs of devices
    known_devices = {
        'F8:1A:67:99:15:E0': 'my super cool gateway at home'
    }
    st.map_devices_arp(
        known_devices=known_devices, ack=True
    )

    it's takes ~10 seconds when a device logs in or out to refresh
    example output with more known devices:
   --------------------------------------------------------------------------------
    arp -a
         IP                MAC           NICKNAME
    192.168.0.2     F8:1A:67:99:15:E0    default gateway Wizzi-Dorms
    192.168.0.5     66:93:D1:80:F9:02
    192.168.0.6     A2:8C:1E:42:57:64    Wizzi-Pixel3a
    192.168.0.7     5C:80:B6:30:1A:7E    Wizzi-Dell (Wireless LAN adapter Wi-Fi)
    192.168.0.10    DC:A6:32:B8:1C:21    rp1-64GB (wlan0)
    192.168.0.8     70:4D:7B:8A:65:EE    Wizzi-Dorms (Ethernet adapter Ethernet)
    --------------------------------------------------------------------------------
    Total run time 0:00:04
    """
    ip_mac = {}
    my_ip = get_active_ipv4()
    if my_ip == 'N/A':
        return ip_mac
    map_brute_force(ack=False)  # refresh arp cache table
    out_lines = mt.run_shell_command_and_get_out(cmd='arp -a', ack_cmd=True, ack_out=False)

    for line in out_lines:
        if 'dynamic' in line:  # break down the output keep only dynamic addresses
            line = ' '.join(line.split())  # many spaces to 1
            line_s = line.split(' ')[:2]  # keep only Internet Address and Physical Address headers
            address, mac = line_s
            ip_mac[address] = mac.replace('-', ':').upper()

    ip_mac[get_active_ipv4()] = get_active_con_mac()  # add this computer
    if ack:
        if ip_mac:
            print(' ' * 4, 'IP', ' ' * 14, 'MAC', ' ' * 9, 'NICKNAME')
        for address, mac in ip_mac.items():
            info = '{:15} {:20}'.format(address, mac)
            if known_devices is not None and mac in known_devices:
                name = known_devices[mac]
                info += ' {}'.format(name)
            print(info)
    return ip_mac


def get_all_network_info(with_colons: bool = True, ack: bool = False) -> dict:
    """
    :param with_colons:
    :param ack:
    :return:
    get all network interfaces (virtual and physical)
    based on https://www.thepythoncode.com/article/get-hardware-system-information-python
    """
    if_addrs = psutil.net_if_addrs()
    info_d = {}
    for interface, interface_addresses in if_addrs.items():
        # print('interface {}:'.format(interface))
        info_d[interface] = {}
        for address in interface_addresses:
            # print('address {}:'.format(interface_addresses))
            k = None
            v = address.address.upper()
            if str(address.family) == 'AddressFamily.AF_INET':
                # from this family we get the ip the interface uses
                k = 'ip'
            elif str(address.family) == 'AddressFamily.AF_INET6':
                # from this family we get the 20 digit mac
                if mt.is_linux():
                    # on linux e.g. address='fe80::8793:7140:756b:1433%wlan0'
                    v = v.split('%')[0].replace('::', ':')
                k = 'mac20' if len(v) == 24 else None  # valid: 20 hexa + 4 semicolons
                if k and not with_colons:
                    v = v.replace(':', '')
            elif str(address.family) in ['AddressFamily.AF_LINK', 'AddressFamily.AF_PACKET']:
                # from this family we get the 12 digit mac (AF_LINK on windows, AF_PACKET on linux)
                k = 'mac12' if len(v) == 17 else None  # valid: 12 hexa + 5 semicolons
                # v is AB-CD-EF-GH-IJ-KL on windows and AB:CD:EF:GH:IJ:KL on linux
                if k:
                    if with_colons:
                        v = v.replace('-', ':')  # handle only windows (linux already in form)
                    else:  # asked to strip delimiter
                        # windows returns xx-yy-... linux returns xx:yy:...
                        v = v.replace('-', '').replace(':', '')  # remove both
            if k:
                info_d[interface][k] = v
    if ack:
        for interface, info in info_d.items():
            print('interface {}:'.format(interface))
            for family_k, address in info.items():
                print('\t{}: {}'.format(family_k, address))

    return info_d


def _get_network_key(interface: str, key: str, with_colons: bool = True) -> str:
    """
    aux function
    :param interface:
    :param key:
    :param with_colons:
    :return:
    """
    mac = None
    info_d = get_all_network_info(with_colons=with_colons, ack=False)
    if interface in info_d:
        if key in info_d[interface]:
            mac = info_d[interface][key]
    return mac


def get_wifi_ipv4(ack: bool = False) -> str:
    """
    :param ack:
    :return: get wifi ip if connected
    """
    interface = None
    if mt.is_windows():
        interface = 'Wi-Fi'
    elif mt.is_linux():
        interface = 'wlan0'
    ip_wifi = _get_network_key(interface=interface, key='ip')
    if ack:
        print('\twifi     ip {}'.format(ip_wifi))
    return ip_wifi


def get_ethernet_ipv4(ack: bool = False) -> str:
    """
    :param ack:
    :return: get ethernet ip if connected
    """
    interface = None
    if mt.is_windows():
        interface = 'Ethernet'
    elif mt.is_linux():
        interface = 'eth0'
    ip_ether = _get_network_key(interface=interface, key='ip')
    if ack:
        print('\tethernet ip {}'.format(ip_ether))
    return ip_ether


def get_wifi_mac(with_colons: bool = True, ack: bool = False) -> str:
    """
    :param with_colons:
    :param ack:
    :return: get wifi mac if connected
    """
    interface = None
    if mt.is_windows():
        interface = 'Wi-Fi'
    elif mt.is_linux():
        interface = 'wlan0'
    mac_wifi = _get_network_key(interface=interface, key='mac12', with_colons=with_colons)
    if ack:
        print('\twifi     mac {}'.format(mac_wifi))
    return mac_wifi


def get_ethernet_mac(with_colons: bool = True, ack: bool = False) -> str:
    """
    :param with_colons:
    :param ack:
    :return: get ethernet mac if connected
    """
    interface = None
    if mt.is_windows():
        interface = 'Ethernet'
    elif mt.is_linux():
        interface = 'eth0'
    mac_ethernet = _get_network_key(interface=interface, key='mac12', with_colons=with_colons)
    if ack:
        print('\tethernet mac {}'.format(mac_ethernet))
    return mac_ethernet


def get_active_con_mac(with_colons: bool = True, ack: bool = False) -> str:
    """
    :param with_colons:
    :param ack:
    :return: mac of wifi/ethernet depends on what is connected and active
    """
    # todo check linux - might be ethe0/wlan0
    active_con = None
    mac12_con = None
    ipv4 = get_active_ipv4()
    if not ipv4 == 'N/A':
        info_d = get_all_network_info(with_colons=with_colons, ack=False)
        for interface, info in info_d.items():
            if 'ip' in info and info['ip'] == ipv4:
                if 'mac12' in info:
                    mac12_con = info['mac12']
                    active_con = interface
        if ack:
            print('\tactive con({}) mac {}'.format(active_con, mac12_con))
    return mac12_con


def get_mac_address_uuid(with_colons: bool = True, ack: bool = False, tabs: int = 1) -> str:
    """
    :return:
    on windows: probably the ethernet mac even if it's disabled
    on linux: active interface mac
    see get_mac_address_test()
    """
    try:
        mac = _get_mac_uuid()
        mac = "{:X}".format(mac)  # turn to Hex
        if with_colons:
            mac = ':'.join(mac[i:i + 2] for i in range(0, 12, 2))
        if ack:
            print('{}* Computer Mac: {}'.format(tabs * '\t', mac))
    except ModuleNotFoundError as e:
        mac = ''
        mt.exception_error(e, real_exception=True)
    return mac


# noinspection PyUnresolvedReferences,PyPackageRequirements
def runQ(query: str, ack_q: bool = True, ack_out: bool = True):  # -> pd.core.frame.DataFrame:
    """
    based on https://cloud.google.com/bigquery/docs/reference/libraries#client-libraries-install-python
    bigQ test online: https://console.cloud.google.com/bigquery
    bigQ generate new key: https://console.cloud.google.com/projectselector2/iam-admin/serviceaccounts
        service and project name > keys > add key > download
    bigQ table keys https://packaging.python.org/en/latest/guides/analyzing-pypi-package-downloads/
    dependencies
    # pip install google # maybe
    # pip install --upgrade google-cloud-speech # maybe
    pip install pandas
    pip install google-cloud-bigquery
    pip install db-dtypes (on administrative mode)

    important: assumes you have credentials file on env variable GOOGLE_APPLICATION_CREDENTIALS
    see big_query_package_by_month_test() to see how do insert dynamically
    """
    from google.cloud import bigquery
    if ack_q:
        print(query)

    client = bigquery.Client()  # Construct a BigQuery client object.
    q_out = client.query(query).to_dataframe()  # Make an API request.

    if ack_out:
        print(q_out)  # as table
    return q_out


# noinspection PyPackageRequirements
def big_query_package_by_month(package_name: str = 'wizzi-utils', months_back: int = 12, only_pip: bool = True):
    """
    querying bigQuery `months_back` on a `package_name`
    bigQuery free usage allows only 1 month diff - so, instead iteratively query 1 month a time and collect result
    :param package_name: e.g. numpy
    :param months_back: how many month back from today
    :param only_pip: if True: only downloads from using pip install. False: all downloads (via bandersnatch)
    :return:
    e.g. output:

        querying last 3 months on package numpy (only_pip True):
        results:
            pip_reqs       month
        0  123265129  2022-12-01
        1  126006551  2022-11-01
        2  129894573  2022-10-01

        querying last 3 months on package numpy (only_pip False):
        results:
                hits       month
        0  129585368  2022-12-01
        1  134565898  2022-11-01
        2  139379190  2022-10-01
        queries time: 0:00:36
    """
    import pandas as pd
    table = 'bigquery-public-data.pypi.file_downloads'
    hits_col_name = 'pip_reqs' if only_pip else 'hits'
    pip_exp = "details.installer.name = 'pip'" if only_pip else "TRUE"

    q_by_months_base = """
                   SELECT
                       COUNT(*) AS {},
                       DATE_TRUNC(DATE(timestamp), MONTH) AS `month`
                   FROM `{}`
                   WHERE
                       file.project = '{}'
                       AND {}
                       AND DATE(timestamp)
                           BETWEEN {}
                           AND {}
                   GROUP BY `month`
                   ORDER BY `month` DESC
               """
    frames = []  # saving each dataframe
    print('querying last {} months on package {} (only_pip {}):'.format(months_back, package_name, only_pip))
    t_all = mt.get_timer()
    for i in range(months_back):
        # t = mt.get_timer()
        date_end = """
                DATE_SUB(
                    DATE_TRUNC(
                        DATE_ADD(
                            DATE_SUB(CURRENT_DATE(), INTERVAL {} MONTH),
                        INTERVAL 1 MONTH),
                    MONTH), 
                INTERVAL 1 DAY)
                """.format(i)
        date_start = 'DATE_TRUNC(DATE_SUB(CURRENT_DATE(), INTERVAL {} MONTH), MONTH)'.format(i)
        q_by_months = q_by_months_base.format(hits_col_name, table, package_name, pip_exp, date_start, date_end)
        df = runQ(query=q_by_months, ack_q=False, ack_out=False)
        frames.append(df)
        # print('\tquery {} done. time passed {}'.format(i + 1, mt.get_timer_delta(s_timer=t)))
    print('results:')
    df_all = pd.concat(frames, ignore_index=True)
    print(df_all)
    print('queries time: {}'.format(mt.get_timer_delta(s_timer=t_all)))
    return
